{-|
Module      : Gargantext.API.GraphQL.Node
Description :
Copyright   : (c) CNRS, 2017
License     : AGPL + CECILL v3
Maintainer  : team@gargantext.org
Stability   : experimental
Portability : POSIX
-}


{-# LANGUAGE DataKinds #-}
{-# LANGUAGE DeriveAnyClass #-}
{-# LANGUAGE DeriveGeneric #-}
{-# LANGUAGE DuplicateRecordFields #-}
{-# LANGUAGE TemplateHaskellQuotes #-}
-- {-# LANGUAGE TypeFamilies #-}

module Gargantext.API.GraphQL.Node where

import Data.Aeson ( Result(..), Value(..) )
import Data.Aeson.KeyMap qualified as KM
import Data.Morpheus.Server.Types (fieldDirective', typeDirective)
import Data.Morpheus.Types ( GQLType(..), Describe(..) )
import Gargantext.API.Admin.Auth.Types ( AuthenticatedUser )
import Gargantext.API.Auth.PolicyCheck ( nodeReadChecks, AccessPolicyManager )
import Gargantext.API.GraphQL.PolicyCheck (withPolicy)
import Gargantext.API.GraphQL.Types ( GqlM, gqlHead )
import Gargantext.Core ( HasDBid(lookupDBid), fromDBid )
import Gargantext.Database.Admin.Types.Node (NodeType)
import Gargantext.Database.Admin.Types.Node qualified as NN
import Gargantext.Database.Prelude (IsDBEnvExtra)  -- , JSONB)
import Gargantext.Database.Query.Table.Node (getDirectChildrenByType, getClosestParentIdByType, getNode)
import Gargantext.Database.Query.Tree qualified as T
import Gargantext.Database.Schema.Node qualified as N
import Gargantext.Prelude
import PUBMED.Types qualified as PUBMED

data Corpus = Corpus
  { id           :: Int
  , name         :: Text
  , parent_id    :: Maybe Int
  , type_id      :: Int
  } deriving (Show, Generic, GQLType)

data Node = Node
  { id           :: Int
  , name         :: Text
  , parent_id    :: Maybe Int
  , type_id      :: Int
  , node_type    :: Maybe NodeType
  } deriving (Show, Eq, Generic)
instance GQLType Node where
  directives _ = typeDirective (Describe "Node is the basic entity type in Gargantext")
instance Ord Node where
  compare (Node { id = id1 }) (Node { id = id2 }) = compare id1 id2



data CorpusByIdArgs
  = CorpusByIdArgs
    { id :: Int
    } deriving (Generic, GQLType)

data NodeByIdArgs
  = NodeByIdArgs
    { id :: Int
    } deriving (Generic, GQLType)

-- | TODO This is not ideal. It would be better to derive
-- NamedResolvers for Node.children, Node.parent
-- https://morpheusgraphql.com/server#named-resolvers
data NodesArgs
  = NodesArgs
    { parentId :: Maybe Int
    , containsChildId :: Maybe Int
    , deepChildId :: Maybe Int
    , nodeType :: Maybe NodeType }
    deriving (Generic)
instance GQLType NodesArgs where
  -- type KIND NodesArgs = TYPE
  directives _ =
    typeDirective (Describe "Arguments to fetch nodes")
    <> fieldDirective' 'parentId Describe { text = "If not null, will fetch nodes having given id as parent (i.e. child nodes)" }
    <> fieldDirective' 'containsChildId Describe { text = "If not null, wil fetch nodes having given id as child (i.e. parent nodes) "}
    <> fieldDirective' 'deepChildId Describe { text = "If not null, will fetch nodes that have given id as a deeply nested child (i.e. ancestor nodes). Can be used to fetch breadcrumbs" }
    <> fieldDirective' 'nodeType Describe { text = "Filters out resulting nodes by given node type" }

-- | Function to resolve user from a query.
getNodeById
  :: (IsDBEnvExtra env)
  => AuthenticatedUser
  -> AccessPolicyManager
  -> NodeByIdArgs
  -> GqlM e env Node
getNodeById autUser mgr NodeByIdArgs { id } =
  withPolicy autUser mgr (nodeReadChecks $ NN.UnsafeMkNodeId id) $ dbNodes id
    >>= gqlHead ("[getNodeById] node not found", id)


getCorpusById
  :: (IsDBEnvExtra env)
  => CorpusByIdArgs -> GqlM e env Corpus
getCorpusById CorpusByIdArgs { id } =
  dbNodesCorpus id
    >>= gqlHead ("[getCorpusById] corpus does not exist", id)

filterNodes
  :: (IsDBEnvExtra env)
  => NodesArgs
  -> GqlM e env [Node]
filterNodes NodesArgs { .. } = do
  children <- case parentId of
    Nothing -> pure []
    Just pId -> dbChildNodes pId nodeType
  parents <- case containsChildId of
    Nothing -> pure []
    Just ccId -> dbParentNodes ccId nodeType
  ancestors <- case deepChildId of
    Nothing -> pure []
    Just dcId -> do
      let nts = maybe NN.allNodeTypes (\nt -> [nt]) nodeType
      ancestors' <- lift $ T.recursiveParents (NN.UnsafeMkNodeId dcId) nts
      pure $ (\(T.DbTreeNode { .. }) -> Node { id = NN.unNodeId _dt_nodeId
                                            , name = _dt_name
                                            , parent_id = NN.unNodeId <$> _dt_parentId
                                            , type_id = _dt_typeId
                                            , node_type = Just $ fromDBid _dt_typeId }) <$> ancestors'
  pure $ children <> parents <> ancestors


dbNodes
  :: (IsDBEnvExtra env)
  => Int -> GqlM e env [Node]
dbNodes node_id = do
  node <- lift $ getNode $ NN.UnsafeMkNodeId node_id
  pure [toNode node]

dbNodesCorpus
  :: (IsDBEnvExtra env)
  => Int -> GqlM e env [Corpus]
dbNodesCorpus corpusId = do
  corpus <- lift $ getNode $ NN.UnsafeMkNodeId corpusId
  pure [toCorpus corpus]

dbParentNodes
  :: (IsDBEnvExtra env)
  => Int
  -> Maybe NodeType
  -> GqlM e env [Node]
dbParentNodes node_id parentType = lift $ do
  -- let mParentType = readEither (T.unpack parent_type) :: Either Prelude.String NodeType
  -- case mParentType of
  --   Left err -> do
  --     lift $ printDebug "[dbParentNodes] error reading parent type" (T.pack err)
  --     pure []
  --   Right parentType -> do
  mNodeId <- getClosestParentIdByType (NN.UnsafeMkNodeId node_id) parentType -- (fromNodeTypeId parent_type_id)
  case mNodeId of
    Nothing -> pure []
    Just id -> do
      node <- getNode id
      pure [toNode node]

dbChildNodes :: (IsDBEnvExtra env)
             => Int
             -> Maybe NodeType
             -> GqlM e env [Node]
dbChildNodes node_id childType = lift $ do
  childIds <- getDirectChildrenByType (NN.UnsafeMkNodeId node_id) childType -- (fromNodeTypeId parent_type_id)
  children <- mapM getNode childIds
  pure $ toNode <$> children

toNode :: NN.Node json -> Node
toNode N.Node { .. } = Node { id = nid
                            , name = _node_name
                            , parent_id = NN.unNodeId <$> _node_parent_id
                            , type_id = _node_typename
                            , node_type = lookupDBid _node_typename
                            }
  where
    nid = NN.unNodeId _node_id

toCorpus :: NN.Node Value -> Corpus
toCorpus N.Node { .. } = Corpus { id = NN.unNodeId _node_id
                                , name = _node_name
                                , parent_id = NN.unNodeId <$> _node_parent_id
                                , type_id = _node_typename }

pubmedAPIKeyFromValue :: Value -> Maybe PUBMED.APIKey
pubmedAPIKeyFromValue (Object kv) =
  case KM.lookup "pubmed_api_key" kv of
    Nothing -> Nothing
    Just v  -> case fromJSON v of
      Error _    -> Nothing
      Success v' -> Just v'
pubmedAPIKeyFromValue _           = Nothing
